# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2022 Mathieu Parent <math.parent@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

from unittest import skipIf
from unittest.mock import call, patch

from packaging import __version__ as packaging_version

from gitlabracadabra.objects.project import GitLabracadabraProject
from gitlabracadabra.tests import my_vcr
from gitlabracadabra.tests.case import TestCaseWithManager


class TestProjectPackageMirrors(TestCaseWithManager):
    """Test package_mirrors param."""

    def assert_has_errors(self, project: GitLabracadabraProject, expected_errors: list[str]) -> None:
        """Asset JSONschema errors.

        Args:
            project: Project.
            expected_errors: List of expected errors (as strings).
        """
        actual_errors = []
        for error in project.errors():
            actual_errors.append(error.message)
        self.assertListEqual(actual_errors, expected_errors)

    def test_package_mirrors_validation(self):
        """Test package_mirrors syntax."""
        project = GitLabracadabraProject('memory', 'group/project', {
            'package_mirrors': [{
                'raw': {
                    'default_url': 'https://example.org/foobar.tgz',
                },
            }],
        })
        self.assert_has_errors(project, [])

        project = GitLabracadabraProject('memory', 'group/project', {
            'package_mirrors': [{
                'enabled': True,
                'raw': {
                    'default_url': 'https://storage.googleapis.com/{package_name}-release/release/{package_version}/bin/linux/amd64/{file_name}',  # noqa: E501
                    'default_package_name': 'kubernetes',
                    'default_package_version': 'v1.20.5',
                    'package_files': [
                        {'url': 'foo'},
                        {'package_name': 'foo'},
                        {'package_version': 'foo'},
                        {'file_name': 'kubectl'},
                        {'url': 'foo', 'package_name': 'foo', 'package_version': 'foo', 'file_name': 'kubectl'},
                    ],
                },
            }],
        })
        self.assert_has_errors(project, [])

        project = GitLabracadabraProject('memory', 'group/project', {
            'package_mirrors': [{
                'raw': {},
            }],
        })
        self.assert_has_errors(project, ["{'raw': {}} is not valid under any of the given schemas"])

        project = GitLabracadabraProject('memory', 'group/project', {
            'package_mirrors': [{
                'github': {
                    'full_name': 'operator-framework/operator-lifecycle-manager',
                    'package_name': 'olm',
                    'tags': ['/v.*/'],
                    'semver': '>=0.18.0',
                    'latest_release': True,
                    'tarball': True,
                    'zipball': True,
                    'assets': ['install.sh', 'crds.yaml', 'olm.yaml'],
                },
            }],
        })
        self.assert_has_errors(project, [])

    @my_vcr.use_cassette
    def test_package_mirrors_raw(self, cass):
        """Test package_mirrors, from raw.

        Args:
            cass: VCR cassette.
        """
        project = GitLabracadabraProject('memory', 'test/test_from_raw', {
            'package_mirrors': [{
                'raw': {
                    'default_url': 'https://download.docker.com/linux/debian/gpg',
                },
            }],
        })
        self.assertEqual(project.errors(), [])
        with patch('gitlabracadabra.packages.destination.logger', autospec=True) as logger:
            project.process()
            self.assertEqual(logger.mock_calls, [
                call.info(
                    '%sUploading %s package file "%s" from "%s" version %s (%s)',
                    '[test/test_from_raw] ',
                    'raw',
                    'gpg',
                    'unknown',
                    '0',
                    'https://download.docker.com/linux/debian/gpg',
                ),
            ])
        self.assertTrue(cass.all_played)

    @my_vcr.use_cassette
    def test_package_mirrors_github(self, cass):
        """Test package_mirrors, from github.

        Args:
            cass: VCR cassette.
        """
        project = GitLabracadabraProject('memory', 'test/test_from_github', {
            'package_mirrors': [{
                'github': {
                    'full_name': 'operator-framework/operator-lifecycle-manager',
                    'package_name': 'olm',
                    'latest_release': True,
                    'assets': ['install.sh', 'Boom'],
                },
            }],
        })
        self.assertEqual(project.errors(), [])
        with patch('gitlabracadabra.packages.destination.logger', autospec=True) as destination_logger:
            with patch('gitlabracadabra.packages.github.logger', autospec=True) as github_logger:
                project.process()
                self.assertEqual(destination_logger.mock_calls, [
                    call.info(
                        '%sUploading %s package file "%s" from "%s" version %s (%s)',
                        '[test/test_from_github] ',
                        'raw',
                        'install.sh',
                        'olm',
                        'v0.18.1',
                        'https://github.com/operator-framework/operator-lifecycle-manager/releases/download/v0.18.1/install.sh',
                    ),
                ])
                self.assertEqual(github_logger.mock_calls, [
                    call.warning(
                        '%sAsset "%s" not found from repository %s in release with tag %s',
                        '[test/test_from_github] ',
                        'Boom',
                        'operator-framework/operator-lifecycle-manager',
                        'v0.18.1',
                    ),
                ])
        self.assertTrue(cass.all_played)

    @my_vcr.use_cassette
    def test_package_mirrors_github_tarball(self, cass):
        """Test package_mirrors, from github tarball: Without Content-Length header.

        Args:
            cass: VCR cassette.
        """
        project = GitLabracadabraProject('memory', 'test/test_from_github', {
            'package_mirrors': [{
                'github': {
                    'full_name': 'projectcalico/calicoctl',
                    'latest_release': True,
                    'tarball': True,
                },
            }],
        })
        self.assertEqual(project.errors(), [])
        with patch('gitlabracadabra.packages.destination.logger', autospec=True) as destination_logger:
            with patch('gitlabracadabra.packages.github.logger', autospec=True) as github_logger:
                project.process()
                self.assertEqual(destination_logger.mock_calls, [
                    call.info(
                        '%sUploading %s package file "%s" from "%s" version %s (%s)',
                        '[test/test_from_github] ',
                        'raw',
                        'calicoctl-v3.18.3.tar.gz',
                        'calicoctl',
                        'v3.18.3',
                        'https://api.github.com/repos/projectcalico/calicoctl/tarball/v3.18.3',
                    ),
                ])
                self.assertEqual(github_logger.mock_calls, [])
        self.assertTrue(cass.all_played)

    @my_vcr.use_cassette(match_on=['method', 'gitlabracadabra_uri', 'gitlabracadabra_body'])
    def test_package_mirrors_helm(self, cass):
        """Test package_mirrors, from helm.

        Args:
            cass: VCR cassette.
        """
        project = GitLabracadabraProject('memory', 'test/test_from_helm', {
            'package_mirrors': [{
                'helm': {
                    'repo_url': 'https://charts.rook.io/release',
                    'package_name': 'rook-ceph',
                },
            }],
        })
        self.assertEqual(project.errors(), [])
        with patch('gitlabracadabra.packages.destination.logger', autospec=True) as destination_logger:
            with patch('gitlabracadabra.packages.helm.logger', autospec=True) as helm_logger:
                project.process()
                self.assertEqual(destination_logger.mock_calls, [
                    call.info(
                        '%sUploading %s package file "%s" from "%s" version %s (%s)',
                        '[test/test_from_helm] ',
                        'helm',
                        'rook-ceph-v1.6.3.tgz',
                        'rook-ceph',
                        'v1.6.3',
                        'https://charts.rook.io/release/rook-ceph-v1.6.3.tgz',
                    ),
                ])
                self.assertEqual(helm_logger.mock_calls, [])
        self.assertTrue(cass.all_played)

    @my_vcr.use_cassette
    def test_package_mirrors_helm_exists(self, cass):
        """Test package_mirrors, from existing helm.

        Args:
            cass: VCR cassette.
        """
        project = GitLabracadabraProject('memory', 'test/test_from_helm', {
            'package_mirrors': [{
                'helm': {
                    'repo_url': 'https://charts.rook.io/release',
                    'package_name': 'rook-ceph',
                },
            }],
        })
        self.assertEqual(project.errors(), [])
        with patch('gitlabracadabra.packages.destination.logger', autospec=True) as destination_logger:
            with patch('gitlabracadabra.packages.helm.logger', autospec=True) as helm_logger:
                project.process()
                self.assertEqual(destination_logger.mock_calls, [])
                self.assertEqual(helm_logger.mock_calls, [])
        self.assertTrue(cass.all_played)

    @skipIf(packaging_version == '19.0', 'PyPI package mirror requires packaging >= 20.9')
    @my_vcr.use_cassette(match_on=['method', 'gitlabracadabra_uri', 'gitlabracadabra_body'])
    def test_package_mirrors_pypi(self, cass):
        """Test package_mirrors, from pypi.

        Args:
            cass: VCR cassette.
        """
        project = GitLabracadabraProject('memory', 'test/test_from_pypi', {
            'package_mirrors': [{
                'pypi': {
                    'requirements': [
                        'types-html5lib',
                        'types-PyYAML',
                    ],
                },
            }],
        })
        self.assertEqual(project.errors(), [])
        with patch('gitlabracadabra.packages.destination.logger', autospec=True) as destination_logger:
            with patch('gitlabracadabra.packages.pypi.logger', autospec=True) as pypi_logger:
                project.process()
                self.assertEqual(destination_logger.mock_calls, [
                    call.info(
                        '%sUploading %s package file "%s" from "%s" version %s (%s)',
                        '[test/test_from_pypi] ',
                        'pypi',
                        'types-html5lib-1.1.0.tar.gz',
                        'types-html5lib',
                        '1.1.0',
                        'https://files.pythonhosted.org/packages/0c/45/2308f303dd263cd1b186ef129cfb20e5938930b96f6ee44293387612ec99/types-html5lib-1.1.0.tar.gz',
                    ),
                    call.info(
                        '%sUploading %s package file "%s" from "%s" version %s (%s)',
                        '[test/test_from_pypi] ',
                        'pypi',
                        'types_html5lib-1.1.0-py3-none-any.whl',
                        'types-html5lib',
                        '1.1.0',
                        'https://files.pythonhosted.org/packages/a9/61/1c6a5536f7ede682e26b9d5a0fe7ba082d3e776bda673cca2db95d7364e0/types_html5lib-1.1.0-py3-none-any.whl',
                    ),
                    call.info(
                        '%sUploading %s package file "%s" from "%s" version %s (%s)',
                        '[test/test_from_pypi] ',
                        'pypi',
                        'types-PyYAML-5.4.3.tar.gz',
                        'types-pyyaml',
                        '5.4.3',
                        'https://files.pythonhosted.org/packages/c1/6a/76f047f9289ed40b8a6460ae9900deec9aa85b9e1276da24112ce8fc9a7b/types-PyYAML-5.4.3.tar.gz',
                    ),
                    call.info(
                        '%sUploading %s package file "%s" from "%s" version %s (%s)',
                        '[test/test_from_pypi] ',
                        'pypi',
                        'types_PyYAML-5.4.3-py2.py3-none-any.whl',
                        'types-pyyaml',
                        '5.4.3',
                        'https://files.pythonhosted.org/packages/b8/86/b957c28571005b1f13b92703400453c6a3a7fddd3f75c4c83bd2e7dc4d47/types_PyYAML-5.4.3-py2.py3-none-any.whl',
                    ),
                ])
                self.assertEqual(pypi_logger.mock_calls, [])
        self.assertTrue(cass.all_played)

    @skipIf(packaging_version == '19.0', 'PyPI package mirror requires packaging >= 20.9')
    @my_vcr.use_cassette(match_on=['method', 'gitlabracadabra_uri', 'gitlabracadabra_body'])
    def test_package_mirrors_pypi_exists(self, cass):
        """Test package_mirrors, from pypi.

        Args:
            cass: VCR cassette.
        """
        project = GitLabracadabraProject('memory', 'test/test_from_pypi', {
            'package_mirrors': [{
                'pypi': {
                    'requirements': [
                        'types-html5lib',
                        'types-PyYAML',
                    ],
                },
            }],
        })
        self.assertEqual(project.errors(), [])
        with patch('gitlabracadabra.packages.destination.logger', autospec=True) as destination_logger:
            with patch('gitlabracadabra.packages.pypi.logger', autospec=True) as pypi_logger:
                project.process()
                self.assertEqual(destination_logger.mock_calls, [])
                self.assertEqual(pypi_logger.mock_calls, [])
        self.assertTrue(cass.all_played)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2022 Mathieu Parent <math.parent@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

from time import time
from typing import TYPE_CHECKING
from urllib.parse import urlparse
from urllib.request import parse_http_list, parse_keqv_list

from requests import PreparedRequest, Response, codes
from requests.structures import CaseInsensitiveDict

from gitlabracadabra import __version__ as gitlabracadabra_version
from gitlabracadabra.auth_info import AuthInfo
from gitlabracadabra.containers.scope import Scope
from gitlabracadabra.session import Session


if TYPE_CHECKING:
    from typing import Any, Iterable, MutableMapping, Optional, Union

    from requests.auth import AuthBase

    Params = Optional[  # noqa: WPS110
        MutableMapping[
            str,
            Union[str, list[str]],
        ]
    ]
    Data = Iterable[bytes]  # noqa: WPS110
    _SimpleParams = dict[str, Union[str, list[str]]]
    _TokenKey = tuple[str, str, Optional[int], Optional[str]]


class Token(object):
    """JWT Token."""

    def __init__(
        self,
        token: str,
        expires_in: int,
    ) -> None:
        """Instantiate a token.

        Args:
            token: Token.
            expires_in: Expires in x seconds.
        """
        minimum_token_lifetime_seconds = 60

        self._token = token
        self._expires_in = expires_in
        if self._expires_in < minimum_token_lifetime_seconds:
            self._expires_in = minimum_token_lifetime_seconds

        # We ignore issued_at property, and use local time instead
        self._issued_at = time()

    @property
    def token(self) -> str:
        """Get token.

        Returns:
            The token.
        """
        return self._token

    @property
    def expiration_time(self) -> float:
        """Get expiration time.

        Returns:
            Expiration time.
        """
        return self._issued_at + self._expires_in

    def is_expired(self) -> bool:
        """Check if token is expired.

        Returns:
            True if token is expired.
        """
        return time() >= self.expiration_time


class AuthenticatedSession(Session):
    """Session with auth per-host."""

    def __init__(self, *args: list[Any], **kwargs: dict[str, Any]) -> None:
        """Instantiate a session.

        Args:
            args: Positional arguments.
            kwargs: Named arguments.
        """
        super().__init__(*args, **kwargs)
        self.headers = CaseInsensitiveDict({
            'User-Agent': 'gitlabracadabra/{0}'.format(gitlabracadabra_version),
            'Docker-Distribution-Api-Version': 'registry/2.0',
        })

        # Added attributes
        self.scheme = 'https'
        self.connection_hostname = ''
        self.auth_info = AuthInfo()
        # Tokens, by set of scheme, host, port and scopes (as query string or None for all scope)
        self._tokens: dict[_TokenKey, Token] = {}
        self._current_scopes: Optional[set[Scope]] = None

    def authenticated_request(
        self,
        method: str,
        url: str,
        params: Optional[Params] = None,  # noqa: WPS110
        data: Optional[Data] = None,  # noqa: WPS110
        headers: Optional[dict[str, str]] = None,
        auth: Optional[AuthBase] = None,
        stream: Optional[bool] = None,
    ) -> Response:
        """Send an HTTP request.

        Args:
            method: HTTP method.
            url: Either a path or a full url.
            params: query string params.
            data: Request body stream.
            headers: Request headers.
            auth: HTTPBasicAuth.
            stream: Stream the response.

        Returns:
            A Response.
        """
        if url.startswith('/'):
            url = '{0}://{1}{2}'.format(self.scheme, self.connection_hostname, url)
        token = self._get_token(url, self._current_scopes)
        if token:
            if headers is None:
                headers = {}
            headers['Authorization'] = 'Bearer {0}'.format(token.token)
        return self.request(
            method,
            url,
            params=params,
            data=data,
            headers=headers,
            auth=auth,
            stream=stream,
        )

    def rebuild_auth(self, prepared_request: PreparedRequest, response: Response) -> None:
        """Override Session method to inject bearer tokens.

        Args:
            prepared_request: Prepared request.
            response: Response.
        """
        super().rebuild_auth(prepared_request, response)  # type: ignore
        token = self._get_token(prepared_request.url or '', self._current_scopes)
        if token:
            prepared_request.headers['Authorization'] = 'Bearer {0}'.format(token.token)

    def connect(self, scopes: Optional[set[Scope]]) -> None:
        """Connect.

        Args:
            scopes: An optional set of scopes.
        """
        self._current_scopes = scopes
        url = '{0}://{1}/v2/'.format(self.scheme, self.connection_hostname)
        token = self._get_token(url, scopes)
        if token:
            return
        token = self._get_token(url, None)
        if token:
            return
        response = self.authenticated_request('get', url)
        if response.history:
            self.connection_hostname = urlparse(response.url).hostname or self.connection_hostname
        if response.status_code == codes['ok']:
            one_hour = 3600
            self._set_token(response, None, Token('no_auth', one_hour))
            return
        if response.status_code == codes['unauthorized']:
            if response.headers['Www-Authenticate'].startswith('Bearer '):
                self._get_bearer_token(response)
                return
        response.raise_for_status()

    def _get_bearer_token(self, response: Response) -> None:
        if self._current_scopes is None:
            raise ValueError
        challenge_parameters = self._get_challenge_parameters(response)
        get_params: _SimpleParams = {}
        if 'service' in challenge_parameters:
            get_params['service'] = challenge_parameters.get('service', 'unknown')
        get_params['scope'] = []
        for scope in sorted(self._current_scopes):
            get_params['scope'].append(  # type: ignore
                'repository:{0}:{1}'.format(scope.remote_name, scope.actions),
            )
        challenge_response = self.authenticated_request(
            'get',
            challenge_parameters['realm'],
            params=get_params,
            headers=self.auth_info.headers,
            auth=self.auth_info.auth,
        )
        challenge_response.raise_for_status()
        json = challenge_response.json()
        self._set_token(
            response,
            self._current_scopes,
            Token(
                str(json.get('token', json.get('access_token', ''))),
                int(json.get('expires_in', 0)),
            ),
        )

    def _get_challenge_parameters(self, response: Response) -> dict[str, str]:
        _, _, challenge = response.headers['Www-Authenticate'].partition('Bearer ')
        return parse_keqv_list(parse_http_list(challenge))

    def _get_token(self, url: str, scopes: Optional[set[Scope]]) -> Optional[Token]:
        parsed = urlparse(url)
        key = (
            parsed.scheme,
            parsed.hostname or '',
            parsed.port,
            self._scopes_hash(scopes),
        )
        token = self._tokens.get(key)
        if token and token.is_expired():
            self._tokens.pop(key)
            return None
        return token

    def _set_token(self, response: Response, scopes: Optional[set[Scope]], token: Token) -> None:
        parsed = urlparse(response.url)
        key = (
            parsed.scheme,
            parsed.hostname or '',
            parsed.port,
            self._scopes_hash(scopes),
        )
        self._tokens[key] = token

    def _scopes_hash(self, scopes: Optional[set[Scope]]) -> Optional[str]:
        if scopes is None:
            return None
        return ','.join(map(str, sorted(scopes)))



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2022 Mathieu Parent <math.parent@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from unittest.mock import call, patch

from gitlabracadabra.packages.helm import Helm
from gitlabracadabra.packages.package_file import PackageFile
from gitlabracadabra.tests import my_vcr
from gitlabracadabra.tests.case import TestCaseWithManager


class TestHelm(TestCaseWithManager):
    """Test Helm class."""

    def test_str(self):
        """Test __str__ method."""
        self.assertEqual(
            str(Helm(log_prefix='foo ', repo_url='https://charts.example.org/', package_name='foobar')),
            'Helm charts repository (url=https://charts.example.org/)',
        )

    @my_vcr.use_cassette
    def test_package_files(self, cass):
        """Test package_files method.

        Args:
            cass: VCR cassette.
        """
        with patch('gitlabracadabra.packages.helm.logger', autospec=True) as logger:
            source = Helm(
                repo_url='https://charts.rook.io/release',
                package_name='rook-ceph',
            )
            self.assertEqual(
                source.package_files, [
                    PackageFile(
                        'https://charts.rook.io/release/rook-ceph-v1.6.3.tgz',
                        'helm',
                        'rook-ceph',
                        'v1.6.3',
                        'rook-ceph-v1.6.3.tgz',
                        metadata={'channel': 'stable'},
                    ),
                ],
            )
            self.assertEqual(logger.mock_calls, [])
        self.assertTrue(cass.all_played)

    @my_vcr.use_cassette
    def test_package_files_index_not_found(self, cass):
        """Test package_files method, with index not found.

        Args:
            cass: VCR cassette.
        """
        with patch('gitlabracadabra.packages.helm.logger', autospec=True) as logger:
            source = Helm(
                repo_url='https://charts.rook.io/not_found',
                package_name='rook-ceph',
            )
            self.assertEqual(source.package_files, [])
            self.assertEqual(logger.mock_calls, [
                call.warning('%sUnexpected HTTP status for Helm index %s: received %i %s', '', 'https://charts.rook.io/not_found/index.yaml', 404, 'Not Found'),
                call.warning('%sPackage not found %s for Helm index %s', '', 'rook-ceph', 'https://charts.rook.io/not_found/index.yaml'),
            ])
        self.assertTrue(cass.all_played)

    @my_vcr.use_cassette
    def test_package_files_relative(self, cass):
        """Test package_files method, with relative urls.

        Args:
            cass: VCR cassette.
        """
        with patch('gitlabracadabra.packages.helm.logger', autospec=True) as logger:
            source = Helm(
                repo_url='https://argoproj.github.io/argo-helm',
                package_name='/ar[g]o/',
            )
            self.assertEqual(
                source.package_files, [
                    PackageFile(
                        'https://argoproj.github.io/argo-helm/argo-1.0.0.tgz',
                        'helm',
                        'argo',
                        '1.0.0',
                        'argo-1.0.0.tgz',
                        metadata={'channel': 'stable'},
                    ),
                ],
            )
            self.assertEqual(logger.mock_calls, [])
        self.assertTrue(cass.all_played)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2022 Mathieu Parent <math.parent@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

from logging import getLogger
from os import getenv
from typing import Optional

from github import Github as PyGithub
from github.GithubException import UnknownObjectException
from github.GitRelease import GitRelease

from gitlabracadabra.matchers import Matcher
from gitlabracadabra.packages.package_file import PackageFile
from gitlabracadabra.packages.source import Source


logger = getLogger(__name__)


class Github(Source):
    """Github source."""

    def __init__(
        self,
        *,
        log_prefix: str = '',
        full_name: str,
        package_name: Optional[str] = None,
        tags: Optional[list[str]] = None,
        semver: Optional[str] = None,
        latest_release: Optional[bool] = None,
        tarball: Optional[bool] = None,
        zipball: Optional[bool] = None,
        assets: Optional[list[str]] = None,
    ) -> None:
        """Initialize a Github source.

        Args:
            log_prefix: Log prefix.
            full_name: Repository full name. Mandatory.
            package_name: Destination package name (defaults to repository name).
            tags: List of tags.
            semver: Semantic version applied on tags.
            latest_release: Get latest release.
            tarball: Get repository tarball (defaults to False).
            zipball: Get repository zip (defaults to False).
            assets: List of assets (None by default).
        """
        self._log_prefix = log_prefix
        self._full_name = full_name
        self._package_name = package_name or full_name.split('/').pop()
        self._tags = tags or []
        self._semver = semver
        self._latest_release = latest_release or False
        self._tarball = tarball or False
        self._zipball = zipball or False
        self._assets = assets or []

        self._repo = PyGithub(
            login_or_token=getenv('GITHUB_TOKEN'),
        ).get_repo(self._full_name, lazy=True)
        self._all_releases: Optional[dict[str, GitRelease]] = None
        self._matching_releases: Optional[dict[str, GitRelease]] = None

    def __str__(self) -> str:
        """Return string representation.

        Returns:
            A string.
        """
        return 'Github repository (full_name={0})'.format(self._full_name)

    @property
    def package_files(self) -> list[PackageFile]:
        """Return list of package files.

        Returns:
            List of package files.
        """
        package_files: list[PackageFile] = []
        for release in self._get_matching_releases().values():
            self._append_package_file_from_release(package_files, release)
        return package_files

    def _get_matching_releases(self) -> dict[str, GitRelease]:
        if self._matching_releases is None:
            matches = Matcher(
                self._tags,
                self._semver,
                log_prefix=self._log_prefix,
            ).match(
                self._get_all_tag_names,
            )
            self._matching_releases = {}
            for match in matches:
                self._append_matching_release(match[0])
            if self._latest_release:
                try:
                    latest_release = self._repo.get_latest_release()
                    self._matching_releases[latest_release.tag_name] = latest_release
                except UnknownObjectException as err2:
                    logger.warning(
                        '%sError getting package files from repository %s, latest release: %s %s',
                        self._log_prefix,
                        self._full_name,
                        err2.status,
                        err2.data.get('message'),
                    )
        return self._matching_releases

    def _get_all_tag_names(self) -> list[str]:
        return list(self._get_all_releases().keys())

    def _get_all_releases(self) -> dict[str, GitRelease]:
        if self._all_releases is None:
            self._all_releases = {}
            for release in self._repo.get_releases():
                self._all_releases[release.tag_name] = release
        return self._all_releases

    def _append_matching_release(self, tag_name: str) -> None:
        if self._all_releases is None:
            try:
                self._matching_releases[tag_name] = self._repo.get_release(tag_name)  # type: ignore
            except UnknownObjectException as err:
                logger.warning(
                    '%sError getting package files from repository %s, release with tag %s: %s %s',
                    self._log_prefix,
                    self._full_name,
                    tag_name,
                    err.status,
                    err.data.get('message'),
                )
        else:
            self._matching_releases[tag_name] = self._all_releases[tag_name]  # type: ignore

    def _append_package_file_from_release(self, package_files: list[PackageFile], release: GitRelease) -> None:

        if self._tarball:
            package_files.append(PackageFile(
                release.tarball_url,
                'raw',
                self._package_name,
                release.tag_name,
                '{0}-{1}.tar.gz'.format(self._full_name.split('/').pop(), release.tag_name),
            ))
        if self._zipball:
            package_files.append(PackageFile(
                release.zipball_url,
                'raw',
                self._package_name,
                release.tag_name,
                '{0}-{1}.zip'.format(self._full_name.split('/').pop(), release.tag_name),
            ))
        if self._assets:
            try:
                # https://github.com/PyGithub/PyGithub/pull/1899
                assets = release.assets  # type: ignore
            except AttributeError:
                assets = release.get_assets()
            assets = {asset.name: asset.browser_download_url for asset in assets}
            for asset_name in self._assets:
                try:
                    package_files.append(PackageFile(
                        assets[asset_name],
                        'raw',
                        self._package_name,
                        release.tag_name,
                        asset_name,
                    ))
                except KeyError:
                    logger.warning(
                        '%sAsset "%s" not found from repository %s in release with tag %s',
                        self._log_prefix,
                        asset_name,
                        self._full_name,
                        release.tag_name,
                    )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2022 Mathieu Parent <math.parent@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

from copy import deepcopy
from logging import getLogger
from typing import Any

from gitlabracadabra.packages.github import Github
from gitlabracadabra.packages.gitlab import Gitlab
from gitlabracadabra.packages.helm import Helm
from gitlabracadabra.packages.pypi import PyPI
from gitlabracadabra.packages.raw import RawSource
from gitlabracadabra.packages.source import Source


logger = getLogger(__name__)


class PackageMirrorsMixin(object):
    """Object (Project) with package mirrors."""

    def _process_package_mirrors(
        self,
        param_name: str,
        param_value: Any,
        dry_run: bool = False,
        skip_save: bool = False,
    ) -> None:
        """Process the package_mirrors param.

        Args:
            param_name: "package_mirrors".
            param_value: List of package mirror dicts.
            dry_run: Dry run.
            skip_save: False.
        """
        assert param_name == 'package_mirrors'  # noqa: S101
        assert not skip_save  # noqa: S101

        destination = Gitlab(
            connection=self.connection,  # type: ignore
            full_path=self._name,  # type: ignore
            project_id=self._obj.id,  # type: ignore
        )

        for package_mirror in param_value:
            if not package_mirror.get('enabled', True):
                continue
            for source_type, source_params in package_mirror.items():
                destination.import_source(
                    self._get_source(source_type, deepcopy(source_params)),
                    dry_run=dry_run,
                )

    def _get_source(self, source_type: str, source_params: dict[str, Any]) -> Source:
        source_class = {
            'raw': RawSource,
            'github': Github,
            'helm': Helm,
            'pypi': PyPI,
        }[source_type]
        source_params['log_prefix'] = '[{0}] '.format(self._name)  # type: ignore
        return source_class(**source_params)  # type: ignore



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2022 Mathieu Parent <math.parent@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from gitlabracadabra.packages.package_file import PackageFile
from gitlabracadabra.tests.case import TestCase


class TestPackageFile(TestCase):
    """Test PackageFile class."""

    def test_attributes(self):
        """Test attributes."""
        package_file = PackageFile(
            'https://example.org/foobar.tgz',
            'raw',
            'my-package-name',
            '1.0',
            'foobar.tar.gz',
            metadata={'hello': 'world'},
        )
        self.assertEqual(package_file.url, 'https://example.org/foobar.tgz')
        self.assertEqual(package_file.package_type, 'raw')
        self.assertEqual(package_file.package_name, 'my-package-name')
        self.assertEqual(package_file.package_version, '1.0')
        self.assertEqual(package_file.file_name, 'foobar.tar.gz')
        self.assertEqual(package_file.metadata, {'hello': 'world'})

    def test_attributes_defaults(self):
        """Test attributes."""
        package_file = PackageFile(
            'https://example.org/foobar.tgz',
            'raw',
            'my-package-name',
        )
        self.assertEqual(package_file.url, 'https://example.org/foobar.tgz')
        self.assertEqual(package_file.package_type, 'raw')
        self.assertEqual(package_file.package_name, 'my-package-name')
        self.assertEqual(package_file.package_version, '0')
        self.assertEqual(package_file.file_name, 'foobar.tgz')
        self.assertEqual(package_file.metadata, {})

    def test_equals(self):
        """Test ==."""
        url = 'https://example.org/foobar.tgz'
        package_file = PackageFile(url, 'raw', 'my-package-name')
        self.assertEqual(
            package_file,
            PackageFile(url, 'raw', 'my-package-name'),
        )
        self.assertEqual(
            package_file,
            PackageFile(url, 'raw', 'my-package-name', '0', 'foobar.tgz', metadata={}),
        )
        self.assertNotEqual(
            package_file,
            PackageFile('!', 'raw', 'my-package-name', '0', 'foobar.tgz', metadata={}),
        )
        self.assertNotEqual(
            package_file,
            PackageFile(url, '!', 'my-package-name', '0', 'foobar.tgz', metadata={}),
        )
        self.assertNotEqual(
            package_file,
            PackageFile(url, 'raw', '!', '0', 'foobar.tgz', metadata={}),
        )
        self.assertNotEqual(
            package_file,
            PackageFile(url, 'raw', 'my-package-name', '!', 'foobar.tgz', metadata={}),
        )
        self.assertNotEqual(
            package_file,
            PackageFile(url, 'raw', 'my-package-name', '0', '!', metadata={}),
        )
        self.assertNotEqual(
            package_file,
            PackageFile(url, 'raw', 'my-package-name', '0', 'foobar.tgz', metadata={'foo': '!'}),
        )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2022 Mathieu Parent <math.parent@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

from typing import TYPE_CHECKING

from requests import Response

from gitlabracadabra.auth_info import AuthInfo
from gitlabracadabra.containers.authenticated_session import AuthenticatedSession
from gitlabracadabra.containers.const import DOCKER_HOSTNAME, DOCKER_REGISTRY
from gitlabracadabra.containers.scope import Scope


if TYPE_CHECKING:
    from typing import Optional

    from requests.auth import AuthBase

    from gitlabracadabra.containers.authenticated_session import Data, Params


class RegistrySession(object):
    """Container registry HTTP methods."""

    def __init__(self, hostname: str, auth_info: Optional[AuthInfo] = None) -> None:
        """Instantiate a registry connection.

        Args:
            hostname: fqdn of a registry.
            auth_info: Authentication information.
        """
        self._session = AuthenticatedSession()
        self._hostname = hostname
        if hostname == DOCKER_HOSTNAME:
            self._session.connection_hostname = DOCKER_REGISTRY
        else:
            self._session.connection_hostname = hostname
        self._session.auth_info = auth_info or AuthInfo()
        # Cache where blobs are present
        # dict key is digest, value is a list of manifest names
        # Used in WithBlobs
        self._blobs: dict[str, list[str]] = {}
        self._sizes: dict[str, int] = {}

    def __del__(self) -> None:  # noqa:WPS603
        """Destroy a registry connection."""
        self._session.close()

    @property
    def hostname(self) -> str:
        """Get hostname.

        Returns:
            The registry hostname.
        """
        return self._hostname

    def request(
        self,
        method: str,
        url: str,
        *,
        scopes: Optional[set[Scope]] = None,
        params: Params = None,  # noqa: WPS110
        data: Optional[Data] = None,  # noqa: WPS110
        headers: Optional[dict[str, str]] = None,
        content_type: Optional[str] = None,
        accept: Optional[tuple[str, ...]] = None,
        auth: Optional[AuthBase] = None,
        stream: Optional[bool] = None,
        raise_for_status: bool = True,
    ) -> Response:
        """Send an HTTP request.

        Args:
            method: HTTP method.
            url: Either a path or a full url.
            scopes: An optional set of scopes.
            params: query string params.
            data: Request body stream.
            headers: Request headers.
            content_type: Uploaded MIME type.
            accept: An optional list of accepted mime-types.
            auth: HTTPBasicAuth.
            stream: Stream the response.
            raise_for_status: Raises `requests.HTTPError`, if one occurred.

        Returns:
            A Response.
        """
        if headers:
            headers = headers.copy()
        else:
            headers = {}
        if accept:
            headers['Accept'] = ', '.join(accept)
        if content_type:
            headers['Content-Type'] = content_type

        self._session.connect(scopes)
        response = self._session.authenticated_request(
            method,
            url,
            params=params,
            data=data,
            headers=headers,
            auth=auth,
            stream=stream,
        )
        if raise_for_status:
            response.raise_for_status()
        return response



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2022 Mathieu Parent <math.parent@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from unittest.mock import call, patch

from gitlabracadabra.packages.github import Github
from gitlabracadabra.packages.package_file import PackageFile
from gitlabracadabra.tests import my_vcr
from gitlabracadabra.tests.case import TestCaseWithManager


class TestGithub(TestCaseWithManager):
    """Test Github class."""

    def test_str(self):
        """Test __str__ method."""
        self.assertEqual(
            str(Github(log_prefix='foo ', full_name='foo/bar')),
            'Github repository (full_name=foo/bar)',
        )

    @my_vcr.use_cassette
    def test_package_files_latest_balls(self, cass):
        """Test package_files method, with latest_release, tarball and zipball.

        Args:
            cass: VCR cassette.
        """
        source = Github(
            full_name='kubernetes-sigs/kubespray',
            latest_release=True,
            tarball=True,
            zipball=True,
        )
        self.assertEqual(
            source.package_files, [
                PackageFile(
                    'https://api.github.com/repos/kubernetes-sigs/kubespray/tarball/v2.15.1',
                    'raw',
                    'kubespray',
                    'v2.15.1',
                    'kubespray-v2.15.1.tar.gz',
                ),
                PackageFile(
                    'https://api.github.com/repos/kubernetes-sigs/kubespray/zipball/v2.15.1',
                    'raw',
                    'kubespray',
                    'v2.15.1',
                    'kubespray-v2.15.1.zip',
                ),
            ],
        )

    @my_vcr.use_cassette
    def test_package_files_assets(self, cass):
        """Test package_files method, with tags, semver and assets.

        Args:
            cass: VCR cassette.
        """
        source = Github(
            log_prefix='[log_prefix] ',
            full_name='operator-framework/operator-lifecycle-manager',
            tags=['/v.*/'],
            semver='>=0.18.0',
            latest_release=True,
            assets=['install.sh', 'crds.yaml', 'olm.yaml', 'unexisting-asset'],
        )
        with patch('gitlabracadabra.packages.github.logger', autospec=True) as logger:
            self.assertEqual(
                source.package_files, [
                    PackageFile(
                        'https://github.com/operator-framework/operator-lifecycle-manager/releases/download/v0.18.1/install.sh',
                        'raw',
                        'operator-lifecycle-manager',
                        'v0.18.1',
                        'install.sh',
                    ),
                    PackageFile(
                        'https://github.com/operator-framework/operator-lifecycle-manager/releases/download/v0.18.1/crds.yaml',
                        'raw',
                        'operator-lifecycle-manager',
                        'v0.18.1',
                        'crds.yaml',
                    ),
                    PackageFile(
                        'https://github.com/operator-framework/operator-lifecycle-manager/releases/download/v0.18.1/olm.yaml',
                        'raw',
                        'operator-lifecycle-manager',
                        'v0.18.1',
                        'olm.yaml',
                    ),
                    PackageFile(
                        'https://github.com/operator-framework/operator-lifecycle-manager/releases/download/v0.18.0/install.sh',
                        'raw',
                        'operator-lifecycle-manager',
                        'v0.18.0',
                        'install.sh',
                    ),
                    PackageFile(
                        'https://github.com/operator-framework/operator-lifecycle-manager/releases/download/v0.18.0/crds.yaml',
                        'raw',
                        'operator-lifecycle-manager',
                        'v0.18.0',
                        'crds.yaml',
                    ),
                    PackageFile(
                        'https://github.com/operator-framework/operator-lifecycle-manager/releases/download/v0.18.0/olm.yaml',
                        'raw',
                        'operator-lifecycle-manager',
                        'v0.18.0',
                        'olm.yaml',
                    ),
                ],
            )
            self.assertEqual(logger.mock_calls, [
                call.warning(
                    '%sAsset "%s" not found from repository %s in release with tag %s',
                    '[log_prefix] ',
                    'unexisting-asset',
                    'operator-framework/operator-lifecycle-manager',
                    'v0.18.1',
                ),
                call.warning(
                    '%sAsset "%s" not found from repository %s in release with tag %s',
                    '[log_prefix] ',
                    'unexisting-asset',
                    'operator-framework/operator-lifecycle-manager',
                    'v0.18.0',
                ),
            ])



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2022 Mathieu Parent <math.parent@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

from gitlabracadabra.packages.package_file import PackageFile


class Source(object):
    """Source package repository."""

    @property
    def package_files(self) -> list[PackageFile]:
        """Return list of package files.

        Raises:
            NotImplementedError: This is an abstract method.
        """
        raise NotImplementedError



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2022 Mathieu Parent <math.parent@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

from re import escape as re_escape
from typing import Optional
from unittest.mock import MagicMock, call, patch

from requests import Response, Session
from requests.auth import AuthBase
from requests.utils import set_environ

from gitlabracadabra.packages.destination import Destination, Stream
from gitlabracadabra.packages.helm import Helm
from gitlabracadabra.packages.package_file import PackageFile
from gitlabracadabra.tests.case import TestCase


class TestDestination(TestCase):
    """Test Destination class."""

    def test_import_source_not_found(self):
        """Test import_source method, with unexisting source."""
        with patch('gitlabracadabra.packages.destination.logger', autospec=True) as logger:
            destination = Destination(log_prefix='[foobar] ')
            source = MagicMock
            source.package_files = [PackageFile('https://source.example.org/not_exists.tgz', 'raw', 'foobar')]
            with patch.object(Session, 'request') as request_mock:
                request_mock.side_effect = self._mocked_request
                destination.import_source(source, dry_run=False)
                self.assertEqual(request_mock.mock_calls, [
                    call('HEAD', 'https://source.example.org/not_exists.tgz'),
                ])
            self.assertEqual(logger.mock_calls, [
                call.warning(
                    '%sNOT uploading %s package file "%s" from "%s" version %s (%s): source not found',
                    '[foobar] ',
                    'raw',
                    'not_exists.tgz',
                    'foobar',
                    '0',
                    'https://source.example.org/not_exists.tgz',
                ),
            ])

    def test_import_destination_exists(self):
        """Test import_source method, with existing destination."""
        with patch('gitlabracadabra.packages.destination.logger', autospec=True) as logger:
            destination = Destination(log_prefix='[foobar] ')
            destination.head_url = MagicMock()
            destination.head_url.return_value = 'https://destination.example.org/foobar.tgz'
            source = MagicMock
            source.package_files = [PackageFile('https://source.example.org/foobar.tgz', 'raw', 'foobar')]
            with patch.object(Session, 'request') as request_mock:
                request_mock.side_effect = self._mocked_request
                destination.import_source(source, dry_run=False)
                self.assertEqual(request_mock.mock_calls, [
                    call('HEAD', 'https://source.example.org/foobar.tgz'),
                    call('HEAD', 'https://destination.example.org/foobar.tgz', headers=None, auth=None),
                ])
            self.assertEqual(logger.mock_calls, [])

    def test_import_dry_run(self):
        """Test import_source method, with dry_run."""
        with patch('gitlabracadabra.packages.destination.logger', autospec=True) as logger:
            destination = Destination(log_prefix='[foobar] ')
            destination.head_url = MagicMock()
            destination.head_url.return_value = 'https://destination.example.org/not_exists.tgz'
            source = MagicMock
            source.package_files = [PackageFile('https://source.example.org/foobar.tgz', 'raw', 'foobar')]
            with patch.object(Session, 'request') as request_mock:
                request_mock.side_effect = self._mocked_request
                destination.import_source(source, dry_run=True)
                self.assertEqual(request_mock.mock_calls, [
                    call('HEAD', 'https://source.example.org/foobar.tgz'),
                    call('HEAD', 'https://destination.example.org/not_exists.tgz', headers=None, auth=None),
                ])
            self.assertEqual(logger.mock_calls, [
                call.info(
                    '%sNOT uploading %s package file "%s" from "%s" version %s (%s): Dry run',
                    '[foobar] ',
                    'raw',
                    'foobar.tgz',
                    'foobar',
                    '0',
                    'https://source.example.org/foobar.tgz',
                ),
            ])

    def test_import_source_proxy_error(self):
        """Test import_source method, with ProxyError raised."""
        with patch('gitlabracadabra.packages.destination.logger', autospec=True) as logger:
            destination = Destination(log_prefix='[foobar] ')
            source = MagicMock
            source.package_files = [PackageFile('https://source.example.org/anything.tgz', 'raw', 'foobar')]
            with set_environ('HTTPS_PROXY', 'http://localhost:42'):
                destination.import_source(source, dry_run=False)
            _, args, _ = logger.mock_calls[0]
            msg = args[8]
            self.assertRegex(msg, re_escape(
                'ProxyError(MaxRetryError("HTTPSConnectionPool(host=\'source.example.org\', port=443): Max retries exceeded with url: /anything.tgz (Caused by ProxyError(\'Cannot connect to proxy.\', NewConnectionError(\'<urllib3.connection.HTTPSConnection object at 0x7f9b3ef3c8e0>: Failed to establish a new connection: [Errno 111] Connection refused\')))',
            ).replace(
                '0x7f9b3ef3c8e0',
                '0x[0-9A-Fa-f]+',
            ).replace(
                r'urllib3\.connection\.HTTPSConnection',
                r'urllib3\.connection\.(Verified)?HTTPSConnection',
            ))
            self.assertEqual(logger.mock_calls, [
                call.warning(
                    '%sError uploading %s package file "%s" from "%s" version %s (%s %s): %s',
                    '[foobar] ',
                    'raw',
                    'anything.tgz',
                    'foobar',
                    '0',
                    'HEAD',
                    'https://source.example.org/anything.tgz',
                    msg,
                ),
            ])

    def test_import_source_proxy_error2(self):
        """Test import_source method, with ProxyError raised while retrieving package files."""
        with patch('gitlabracadabra.packages.destination.logger', autospec=True) as logger:
            destination = Destination(log_prefix='[foobar] ')
            source = Helm(log_prefix='[foobar] ', repo_url='https://source.example.org', package_name='foo')
            with set_environ('HTTPS_PROXY', 'http://localhost:42'):
                destination.import_source(source, dry_run=False)
            _, args, _ = logger.mock_calls[0]
            msg = args[5]
            self.assertRegex(msg, re_escape(
                'ProxyError(MaxRetryError("HTTPSConnectionPool(host=\'source.example.org\', port=443): Max retries exceeded with url: /index.yaml (Caused by ProxyError(\'Cannot connect to proxy.\', NewConnectionError(\'<urllib3.connection.HTTPSConnection object at 0x7f9b3ef3c8e0>: Failed to establish a new connection: [Errno 111] Connection refused\')))',
            ).replace(
                '0x7f9b3ef3c8e0',
                '0x[0-9A-Fa-f]+',
            ).replace(
                r'urllib3\.connection\.HTTPSConnection',
                r'urllib3\.connection\.(Verified)?HTTPSConnection',
            ))
            self.assertEqual(logger.mock_calls, [
                call.warning(
                    '%sError retrieving package files list from %s (%s %s): %s',
                    '[foobar] ',
                    str(source),
                    'GET',
                    'https://source.example.org/index.yaml',
                    msg,
                ),
            ])

    def test_import_upload(self):
        """Test import_source method, without dry_run."""
        with patch('gitlabracadabra.packages.destination.logger', autospec=True) as logger:
            destination = Destination(log_prefix='[foobar] ')
            destination.head_url = MagicMock()
            destination.head_url.return_value = 'https://destination.example.org/not_exists.tgz'
            source = MagicMock
            source.package_files = [PackageFile('https://source.example.org/foobar.tgz', 'raw', 'foobar')]
            with patch.object(Session, 'request') as request_mock:
                request_mock.side_effect = self._mocked_request
                destination.import_source(source, dry_run=False)
                with patch.object(Stream, '__eq__') as stream_eq_mock:
                    stream_eq_mock.return_value = True
                    self.assertEqual(request_mock.mock_calls, [
                        call('HEAD', 'https://source.example.org/foobar.tgz'),
                        call('HEAD', 'https://destination.example.org/not_exists.tgz', headers=None, auth=None),
                        call('GET', 'https://source.example.org/foobar.tgz', stream=True, headers={'Accept-Encoding': '*'}),
                        call('PUT', 'https://destination.example.org/not_exists.tgz', data=Stream('a file-like object'), headers=None, auth=None),  # noqa: E501
                    ])
            self.assertEqual(logger.mock_calls, [
                call.info(
                    '%sUploading %s package file "%s" from "%s" version %s (%s)',
                    '[foobar] ',
                    'raw',
                    'foobar.tgz',
                    'foobar',
                    '0',
                    'https://source.example.org/foobar.tgz',
                ),
            ])

    def _mocked_request(
        self,
        method: str,
        url: str,
        data: Optional[Stream] = None,  # noqa: WPS110
        headers: Optional[dict[str, str]] = None,
        stream: Optional[bool] = None,
        auth: Optional[AuthBase] = None,
    ) -> Response:
        response = Response()
        if method in {'HEAD', 'GET'}:
            if url in {'https://source.example.org/foobar.tgz', 'https://destination.example.org/foobar.tgz'}:
                response.status_code = 200
                if stream is True:
                    response.raw = 'a file-like object'
            else:
                response.status_code = 404
        elif method == 'PUT':
            if url == 'https://destination.example.org/not_exists.tgz':
                response.status_code = 201
        return response



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2022 Mathieu Parent <math.parent@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from gitlabracadabra.packages.package_file import PackageFile
from gitlabracadabra.packages.raw import RawSource
from gitlabracadabra.tests.case import TestCase


class TestRawSource(TestCase):
    """Test RawSource class."""

    def test_str(self):
        """Test __str__ method."""
        self.assertEqual(
            str(RawSource(log_prefix='foo ', default_url='https://foobar.example.org')),
            'Raw repository (default_url=https://foobar.example.org)',
        )

    def test_package_files_mono(self):
        """Test package_files method, with only mandatory arguments."""
        source = RawSource(
            default_url='https://example.org/foobar.tgz',
        )
        self.assertEqual(
            source.package_files,
            [PackageFile('https://example.org/foobar.tgz', 'raw', 'unknown', '0', 'foobar.tgz')],
        )

    def test_package_files_format(self):
        """Test package_files method, with format in url."""
        source = RawSource(
            default_url='https://example.org/{package_name}/{package_version}/pkg.tgz',
            package_files=[
                {'package_name': 'pkg1', 'package_version': '2.0', 'file_name': 'file_name.tgz'},
                {'package_name': 'pkg2', 'package_version': '3.0'},
                {'file_name': 'file_name2.tgz'},
                {},
            ],
        )
        self.assertEqual(
            source.package_files,
            [
                PackageFile('https://example.org/pkg1/2.0/pkg.tgz', 'raw', 'pkg1', '2.0', 'file_name.tgz'),
                PackageFile('https://example.org/pkg2/3.0/pkg.tgz', 'raw', 'pkg2', '3.0', 'pkg.tgz'),
                PackageFile('https://example.org/unknown/0/pkg.tgz', 'raw', 'unknown', '0', 'file_name2.tgz'),
                PackageFile('https://example.org/unknown/0/pkg.tgz', 'raw', 'unknown', '0', 'pkg.tgz'),
            ],
        )

    def test_package_files_format_file_name(self):
        """Test package_files method, with default_* and file_name."""
        source = RawSource(
            default_url='https://storage.googleapis.com/{package_name}-release/release/{package_version}/bin/linux/amd64/{file_name}',  # noqa: E501
            default_package_name='kubernetes',
            default_package_version='v1.20.5',
            package_files=[
                {'file_name': 'kubectl'},
                {'file_name': 'kubelet'},
                {'file_name': 'kubeadm'},
            ],
        )
        self.assertEqual(
            source.package_files,
            [
                PackageFile('https://storage.googleapis.com/kubernetes-release/release/v1.20.5/bin/linux/amd64/kubectl', 'raw', 'kubernetes', 'v1.20.5', 'kubectl'),  # noqa: E501
                PackageFile('https://storage.googleapis.com/kubernetes-release/release/v1.20.5/bin/linux/amd64/kubelet', 'raw', 'kubernetes', 'v1.20.5', 'kubelet'),  # noqa: E501
                PackageFile('https://storage.googleapis.com/kubernetes-release/release/v1.20.5/bin/linux/amd64/kubeadm', 'raw', 'kubernetes', 'v1.20.5', 'kubeadm'),  # noqa: E501
            ],
        )

    def test_package_files_format_url(self):
        """Test package_files method, with default_* and url."""
        source = RawSource(
            default_url='https://storage.googleapis.com/{package_name}-release/release/{package_version}/bin/linux/amd64',  # noqa: E501
            default_package_name='kubernetes',
            default_package_version='v1.20.5',
            package_files=[
                {'url': '{default_url}/kubectl'},
                {'url': '{default_url}/kubelet'},
                {'url': '{default_url}/kubeadm'},
            ],
        )
        self.assertEqual(
            source.package_files,
            [
                PackageFile('https://storage.googleapis.com/kubernetes-release/release/v1.20.5/bin/linux/amd64/kubectl', 'raw', 'kubernetes', 'v1.20.5', 'kubectl'),  # noqa: E501
                PackageFile('https://storage.googleapis.com/kubernetes-release/release/v1.20.5/bin/linux/amd64/kubelet', 'raw', 'kubernetes', 'v1.20.5', 'kubelet'),  # noqa: E501
                PackageFile('https://storage.googleapis.com/kubernetes-release/release/v1.20.5/bin/linux/amd64/kubeadm', 'raw', 'kubernetes', 'v1.20.5', 'kubeadm'),  # noqa: E501
            ],
        )

    def test_package_files_format_edge_cases(self):
        """Test package_files method, with {file_name} in url."""
        source = RawSource(
            default_url='http://example.org/{default_package_name}/{package_name}/{default_package_version}/{package_version}/{file_name}',  # noqa: E501
            package_files=[
                {'url': '{default_url}/foo', 'package_name': 'pkg2', 'package_version': '3.0'},
                {},
            ],
        )
        self.assertEqual(
            source.package_files,
            [
                PackageFile('http://example.org/unknown/pkg2/0/3.0/{file_name}/foo', 'raw', 'pkg2', '3.0', 'foo'),
                PackageFile('http://example.org/unknown/unknown/0/0/{file_name}', 'raw', 'unknown', '0', '{file_name}'),
            ],
        )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2022 Mathieu Parent <math.parent@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

from logging import getLogger
from typing import TYPE_CHECKING, Optional
from urllib.parse import quote

from gitlabracadabra.auth_info import AuthInfo
from gitlabracadabra.packages.destination import Destination
from gitlabracadabra.packages.package_file import PackageFile


if TYPE_CHECKING:
    from gitlabracadabra.gitlab.connection import GitlabConnection


HELM = 'helm'
PYPI = 'pypi'

logger = getLogger(__name__)


class Gitlab(Destination):
    """Gitlab repository."""

    def __init__(
        self,
        *,
        connection: GitlabConnection,
        full_path: str,
        project_id: int,
    ) -> None:
        """Initialize Gitlab repository.

        Args:
            connection: A Gitlab connection.
            full_path: Project full path.
            project_id: Project ID.
        """
        self._connection = connection
        self._full_path = full_path
        self._project_id = project_id
        super().__init__(log_prefix='[{0}] '.format(full_path))

    def upload_method(self, package_file: PackageFile) -> str:
        """Get upload HTTP method.

        Args:
            package_file: Source package file.

        Returns:
            The upload method.
        """
        if package_file.package_type in {HELM, PYPI}:
            return 'POST'

        return super().upload_method(package_file)

    def head_url(self, package_file: PackageFile) -> str:
        """Get URL to test existence of destination package file with a HEAD request.

        Args:
            package_file: Source package file.

        Returns:
            An URL.

        Raises:
            NotImplementedError: For unsupported package types.
        """
        if package_file.package_type == 'raw':
            return '{0}/projects/{1}/packages/generic/{2}/{3}/{4}'.format(
                self._connection.api_url,
                quote(self._full_path, safe=''),
                quote(package_file.package_name, safe=''),  # [A-Za-z0-9\.\_\-\+]+
                quote(package_file.package_version, safe=''),  # (\.?[\w\+-]+\.?)+
                quote(package_file.file_name, safe=''),  # [A-Za-z0-9\.\_\-\+]+
            )
        elif package_file.package_type == HELM:
            channel = package_file.metadata.get('channel') or 'stable'
            file_name = '{0}-{1}.tgz'.format(package_file.package_name, package_file.package_version)
            return '{0}/projects/{1}/packages/helm/{2}/charts/{3}'.format(
                self._connection.api_url,
                quote(self._full_path, safe=''),
                quote(channel, safe=''),
                quote(file_name, safe=''),
            )
        elif package_file.package_type == PYPI:
            return '{0}/projects/{1}/packages/pypi/files/{2}/{3}'.format(
                self._connection.api_url,
                self._project_id,
                quote(package_file.metadata.get('sha256', ''), safe=''),
                quote(package_file.file_name, safe=''),
            )

        raise NotImplementedError

    def upload_url(self, package_file: PackageFile) -> str:
        """Get URL to upload to.

        Args:
            package_file: Source package file.

        Returns:
            The upload URL.
        """
        if package_file.package_type == HELM:
            channel = package_file.metadata.get('channel') or 'stable'
            return '{0}/projects/{1}/packages/helm/api/{2}/charts'.format(
                self._connection.api_url,
                quote(self._full_path, safe=''),
                quote(channel, safe=''),
            )
        elif package_file.package_type == PYPI:
            return (
                '{0}/projects/{1}/packages/pypi?' +
                'requires_python={2}&' +
                'name={3}&' +
                'version={4}&' +
                'md5_digest={5}&' +
                'sha256_digest={6}'
            ).format(
                self._connection.api_url,
                self._project_id,
                quote(package_file.metadata.get('requires-python', ''), safe=''),
                quote(package_file.package_name, safe=''),
                quote(package_file.package_version, safe=''),
                quote(package_file.metadata.get('md5', ''), safe=''),
                quote(package_file.metadata.get('sha256', ''), safe=''),
            )

        return super().upload_url(package_file)

    def files_key(self, package_file: PackageFile) -> Optional[str]:
        """Get files key, to upload to. If None, uploaded as body.

        Args:
            package_file: Source package file.

        Returns:
            The files key, or None.
        """
        if package_file.package_type == HELM:
            return 'chart'
        elif package_file.package_type == PYPI:
            return 'content'

        return super().files_key(package_file)

    def auth_info(self, package_file: PackageFile) -> AuthInfo:
        """Get auth info when testing existence and uploading.

        Args:
            package_file: Source package file.

        Returns:
            An AuthInfo.
        """
        return self._connection.registry_auth_info



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2022 Mathieu Parent <math.parent@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from unittest import skipIf

from gitlab import __version__ as gitlab_version

from packaging.version import parse as version_parse

from gitlabracadabra.objects.application_settings import GitLabracadabraApplicationSettings
from gitlabracadabra.tests import my_vcr
from gitlabracadabra.tests.case import TestCaseWithManager


class TestApplicationSettings(TestCaseWithManager):
    @skipIf(version_parse(gitlab_version) < version_parse('3.7.0'), 'python-gitlab without proper array support')
    @my_vcr.use_cassette
    def test_settings(self, cass):
        # On the GitLab rails console, before recording:
        # ApplicationSetting.first.delete
        obj = GitLabracadabraApplicationSettings('memory', 'application_settings', {
            # 'abuse_notification_email': 'some string',
            # 'admin_mode': True,
            'after_sign_out_path': 'https://example.org/sign_out',
            'after_sign_up_text': 'some string',
            'akismet_api_key': 'some string',
            'akismet_enabled': True,
            'allow_group_owners_to_manage_ldap': False,
            # 'allow_local_requests_from_system_hooks': False,
            # 'allow_local_requests_from_web_hooks_and_services': True,
            'archive_builds_in_human_readable': '15 days',
            # 'asset_proxy_allowlist': 'some string',
            # 'asset_proxy_enabled': True,
            # 'asset_proxy_secret_key': 'some string',
            # 'asset_proxy_url': 'some string',
            'authorized_keys_enabled': False,
            'auto_devops_domain': 'autodevops.example.org',
            'auto_devops_enabled': True,
            # 'automatic_purchased_storage_allocation': True,
            'check_namespace_plan': True,
            'commit_email_hostname': 'users.noreply.127.0.0.1',
            # 'container_expiration_policies_enable_historic_entries': True,
            # 'container_registry_cleanup_tags_service_max_list_size': 200,
            # 'container_registry_delete_tags_service_timeout': 250,
            # 'container_registry_expiration_policies_caching': False,
            # 'container_registry_expiration_policies_worker_capacity': 4,
            # 'container_registry_import_created_before': '2022-01-23 00:00:00',
            # 'container_registry_import_max_retries': 3,
            # 'container_registry_import_max_step_duration': 300,
            # 'container_registry_import_max_tags_count': 100,
            # 'container_registry_import_start_max_retries': 50,
            # 'container_registry_import_target_plan': 'free',
            'container_registry_token_expire_delay': 5,
            # 'custom_http_clone_url_root': 'some string',
            # 'deactivate_dormant_users': 'some string',
            'default_artifacts_expire_in': '30 days',
            # 'default_branch_name': 'some string',
            'default_branch_protection': 2,
            'default_ci_config_path': 'some string',
            'default_group_visibility': 'internal',
            'default_project_creation': 2,
            # 'default_project_deletion_protection': True,
            'default_project_visibility': 'internal',
            'default_projects_limit': 100000,
            'default_snippet_visibility': 'internal',
            # 'delayed_project_deletion': 'some string',
            # 'delete_inactive_projects': 'some string',
            # 'deletion_adjourned_period': 7,
            # 'diff_max_files': 1000,
            # 'diff_max_lines': 50000,
            'diff_max_patch_bytes': 204800,
            # 'disable_feed_token': True,
            # 'disable_overriding_approvers_per_merge_request': 'some string',
            'disabled_oauth_sign_in_sources': [],
            'dns_rebinding_protection_enabled': False,
            # 'domain_allowlist': [],
            # 'domain_allowlist_raw': 'some string',
            # 'domain_denylist': 'some string',
            # 'domain_denylist_enabled': True,
            # 'domain_denylist_raw': 'some string',
            'dsa_key_restriction': 0,
            'ecdsa_key_restriction': 0,
            # 'ecdsa_sk_key_restriction': 0,
            'ed25519_key_restriction': 0,
            # 'ed25519_sk_key_restriction': 0,
            # 'eks_access_key_id': 'some string',
            # 'eks_account_id': 'some string',
            # 'eks_integration_enabled': True,
            # 'eks_secret_access_key': 'some string',
            # 'elasticsearch_analyzers_kuromoji_enabled': True,
            # 'elasticsearch_analyzers_kuromoji_search': True,
            # 'elasticsearch_analyzers_smartcn_enabled': True,
            # 'elasticsearch_analyzers_smartcn_search': True,
            'elasticsearch_aws': True,
            'elasticsearch_aws_access_key': 'some string',
            'elasticsearch_aws_region': 'us-east-1',
            'elasticsearch_aws_secret_access_key': 'some string',
            # 'elasticsearch_client_request_timeout': 0,
            # 'elasticsearch_indexed_field_length_limit': 0,
            # 'elasticsearch_indexed_file_size_limit_kb': 1024,
            'elasticsearch_indexing': True,
            'elasticsearch_limit_indexing': True,
            # 'elasticsearch_max_bulk_concurrency': 10,
            # 'elasticsearch_max_bulk_size_mb': 'some string',
            'elasticsearch_namespace_ids': [1],
            # 'elasticsearch_password': 'some string',
            # 'elasticsearch_pause_indexing': 'some string',
            'elasticsearch_project_ids': [2],
            # 'elasticsearch_replicas': 'some string',
            'elasticsearch_search': True,
            # 'elasticsearch_shards': 'some string',
            'elasticsearch_url': 'http://localhost:9200',
            # 'elasticsearch_username': 'some string',
            # 'email_additional_text': 'some string',  # current value is not available
            'email_author_in_body': True,
            # 'email_restrictions': 'some string',
            # 'email_restrictions_enabled': True,
            'enabled_git_access_protocol': 'http',
            # 'enforce_namespace_storage_limit': True,
            # 'enforce_pat_expiration': False,
            # 'enforce_ssh_key_expiration': 'some string',
            # 'enforce_terms': True,
            # 'external_auth_client_cert': 'some string',  # FIXME is not a valid X509 certificate.
            # 'external_auth_client_key': 'some string',  # FIXME could not read private key, is the passphrase correct?
            'external_auth_client_key_pass': 'some string',
            'external_authorization_service_default_label': 'some string',
            'external_authorization_service_enabled': True,
            'external_authorization_service_timeout': 5,
            'external_authorization_service_url': 'https://example.org/external_authorization_service_url',
            # 'external_pipeline_validation_service_timeout': 'some string',
            # 'external_pipeline_validation_service_token': 'some string',
            # 'external_pipeline_validation_service_url': 'some string',
            # 'file_template_project_id': 42,  # current value is not available
            'first_day_of_week': 'monday',
            # 'floc_enabled': True,
            # 'force_pages_access_control': 'some string',
            'geo_node_allowed_ips': '0.0.0.0/0, ::/0',
            'geo_status_timeout': 10,
            # 'git_two_factor_session_expiry': 15,
            'gitaly_timeout_default': 57,
            'gitaly_timeout_fast': 20,
            'gitaly_timeout_medium': 56,
            # 'gitpod_enabled': True,
            # 'gitpod_url': 'https://gitpod.io/',
            'grafana_enabled': True,
            'grafana_url': 'some string',
            'gravatar_enabled': False,
            # 'group_download_export_limit': 1,
            # 'group_export_limit': 6,
            # 'group_import_limit': 6,
            # 'group_owners_can_manage_default_branch_protection': 'some string',
            # 'group_runner_token_expiration_interval': 'some string',
            'hashed_storage_enabled': True,
            # 'help_page_documentation_base_url': 'some string',
            'help_page_hide_commercial_content': True,
            'help_page_support_url': 'https://example.org/help_page_support_url',
            'help_page_text': 'some string',
            'help_text': 'some string',
            'hide_third_party_offers': True,
            'home_page_url': 'https://example.org/home_page_url',
            'housekeeping_enabled': False,
            'housekeeping_full_repack_period': 50,
            'housekeeping_gc_period': 200,
            'housekeeping_incremental_repack_period': 10,
            'html_emails_enabled': True,
            'import_sources': [
                'github',
                'bitbucket',
                'bitbucket_server',
                'gitlab',
                # 'google_code',
                'fogbugz',
                'git',
                'gitlab_project',
                'gitea',
                'manifest',
                'phabricator',
            ],
            # 'in_product_marketing_emails_enabled': True,
            # 'inactive_projects_delete_after_months': 'some string',
            # 'inactive_projects_min_size_mb': 'some string',
            # 'inactive_projects_send_warning_email_after_months': 'some string',
            # 'invisible_captcha_enabled': True,
            # 'issues_create_limit': 300,
            # 'keep_latest_artifact': 'some string',
            # 'kroki_enabled': True,
            # 'kroki_formats': {'blockdiag': True, 'bpmn': True, 'excalidraw': True},
            # 'kroki_url': 'some string',
            'local_markdown_version': 0,
            # 'lock_memberships_to_ldap': True,
            # 'login_recaptcha_protection_enabled': True,
            # 'mailgun_events_enabled': True,
            # 'mailgun_signing_key': 'some string',
            # 'maintenance_mode': True,
            # 'maintenance_mode_message': 'some string',
            'max_artifacts_size': 100,
            'max_attachment_size': 10,
            # 'max_import_size': 0,
            'max_pages_size': 12,
            # 'max_personal_access_token_lifetime': 'some string',
            # 'max_ssh_key_lifetime': 'some string',
            # 'max_yaml_depth': 100,
            # 'max_yaml_size_bytes': 1048576,
            'metrics_method_call_threshold': 42,
            # 'minimum_password_length': 8,
            'mirror_available': False,
            # 'mirror_capacity_threshold': 15,  # current value is not available
            # 'mirror_max_capacity': 30,  # current value is not available
            # 'mirror_max_delay': 300,  # current value is not available
            # 'notes_create_limit': 300,
            # 'notes_create_limit_allowlist_raw': 'some string',
            # 'notify_on_unknown_sign_in': False,
            # 'npm_package_requests_forwarding': 'some string',
            # 'outbound_local_requests_allowlist_raw': 'some string',
            'pages_domain_verification_enabled': True,
            # 'password_authentication_enabled': True,
            # 'password_authentication_enabled_for_git': False,
            # 'password_authentication_enabled_for_web': False,
            # 'performance_bar_allowed_group_path': 'some string',  # current value is not available
            # 'personal_access_token_prefix': 'some string',
            'plantuml_enabled': True,
            'plantuml_url': 'some string',
            'polling_interval_multiplier': 1,
            # 'prevent_merge_requests_author_approval': 'some string',
            # 'prevent_merge_requests_committers_approval': 'some string',
            # 'project_download_export_limit': 1,
            'project_export_enabled': False,
            # 'project_export_limit': 6,
            # 'project_import_limit': 6,
            # 'project_runner_token_expiration_interval': 'some string',
            'prometheus_metrics_enabled': True,
            'protected_ci_variables': False,
            # 'protected_paths_raw': 'some string',
            'pseudonymizer_enabled': True,
            # 'push_event_activities_limit': 3,
            # 'push_event_hooks_limit': 3,
            # 'pypi_package_requests_forwarding': 'some string',
            # 'rate_limiting_response_text': 'some string',
            # 'raw_blob_request_limit': 300,
            'recaptcha_enabled': True,
            'recaptcha_private_key': 'some string',
            'recaptcha_site_key': 'some string',
            'receive_max_input_size': 42,
            'repository_checks_enabled': False,
            'repository_size_limit': 42,
            # 'repository_storages_weighted': {'default': 100},
            # 'require_admin_approval_after_user_signup': False,
            'require_two_factor_authentication': True,
            'restricted_visibility_levels': [],
            'rsa_key_restriction': 0,
            # 'runner_token_expiration_interval': 'some string',
            # 'search_rate_limit': 30,
            # 'search_rate_limit_unauthenticated': 10,
            # 'secret_detection_revocation_token_types_url': 'some string',
            # 'secret_detection_token_revocation_enabled': True,
            # 'secret_detection_token_revocation_token': 'some string',
            # 'secret_detection_token_revocation_url': 'some string',
            'send_user_confirmation_email': True,
            # 'sentry_clientside_dsn': 'some string',
            # 'sentry_dsn': 'some string',
            # 'sentry_enabled': True,
            # 'sentry_environment': 'some string',
            'session_expire_delay': 10080,
            'shared_runners_enabled': False,
            'shared_runners_minutes': 42,
            'shared_runners_text': 'some string',
            # 'sidekiq_job_limiter_compression_threshold_bytes': 100000,
            # 'sidekiq_job_limiter_limit_bytes': 0,
            # 'sidekiq_job_limiter_mode': 'compress',
            'sign_in_text': 'some string',
            # 'signin_enabled': True,
            'signup_enabled': False,
            'slack_app_enabled': True,
            'slack_app_id': 'some string',
            'slack_app_secret': 'some string',
            'slack_app_verification_token': 'some string',
            # 'snippet_size_limit': 52428800,
            # 'snowplow_app_id': 'some string',
            # 'snowplow_collector_hostname': 'some string',
            # 'snowplow_cookie_domain': 'some string',
            # 'snowplow_enabled': True,
            # 'sourcegraph_enabled': True,
            # 'sourcegraph_public_only': False,
            # 'sourcegraph_url': 'some string',
            # 'spam_check_api_key': 'some string',
            # 'spam_check_endpoint_enabled': True,
            # 'spam_check_endpoint_url': 'some string',
            # 'static_objects_external_storage_auth_token': 'some string',
            # 'static_objects_external_storage_url': 'some string',
            # 'suggest_pipeline_enabled': False,
            'terminal_max_session_time': 0,
            'terms': 'some string',
            'throttle_authenticated_api_enabled': True,
            'throttle_authenticated_api_period_in_seconds': 3600,
            'throttle_authenticated_api_requests_per_period': 7200,
            # 'throttle_authenticated_deprecated_api_enabled': True,
            # 'throttle_authenticated_deprecated_api_period_in_seconds': 3600,
            # 'throttle_authenticated_deprecated_api_requests_per_period': 3600,
            # 'throttle_authenticated_files_api_enabled': True,
            # 'throttle_authenticated_files_api_period_in_seconds': 15,
            # 'throttle_authenticated_files_api_requests_per_period': 500,
            # 'throttle_authenticated_git_lfs_enabled': True,
            # 'throttle_authenticated_git_lfs_period_in_seconds': 60,
            # 'throttle_authenticated_git_lfs_requests_per_period': 1000,
            # 'throttle_authenticated_packages_api_enabled': True,
            # 'throttle_authenticated_packages_api_period_in_seconds': 15,
            # 'throttle_authenticated_packages_api_requests_per_period': 1000,
            'throttle_authenticated_web_enabled': True,
            'throttle_authenticated_web_period_in_seconds': 3600,
            'throttle_authenticated_web_requests_per_period': 7200,
            # 'throttle_incident_management_notification_enabled': True,
            # 'throttle_incident_management_notification_per_period': 3600,
            # 'throttle_incident_management_notification_period_in_seconds': 3600,
            # 'throttle_protected_paths_enabled': True,
            # 'throttle_protected_paths_period_in_seconds': 'some string',
            # 'throttle_protected_paths_requests_per_period': 'some string',
            # 'throttle_unauthenticated_api_enabled': True,
            # 'throttle_unauthenticated_api_period_in_seconds': 3600,
            # 'throttle_unauthenticated_api_requests_per_period': 3600,
            # 'throttle_unauthenticated_deprecated_api_enabled': True,
            # 'throttle_unauthenticated_deprecated_api_period_in_seconds': 3600,
            # 'throttle_unauthenticated_deprecated_api_requests_per_period': 1800,
            'throttle_unauthenticated_enabled': True,
            # 'throttle_unauthenticated_files_api_enabled': True,
            # 'throttle_unauthenticated_files_api_period_in_seconds': 15,
            # 'throttle_unauthenticated_files_api_requests_per_period': 125,
            # 'throttle_unauthenticated_packages_api_enabled': True,
            # 'throttle_unauthenticated_packages_api_period_in_seconds': 15,
            # 'throttle_unauthenticated_packages_api_requests_per_period': 800,
            'throttle_unauthenticated_period_in_seconds': 3600,
            'throttle_unauthenticated_requests_per_period': 3600,
            # 'throttle_unauthenticated_web_enabled': True,
            # 'throttle_unauthenticated_web_period_in_seconds': 'some string',
            # 'throttle_unauthenticated_web_requests_per_period': 'some string',
            'time_tracking_limit_to_hours': True,
            'two_factor_grace_period': 48,
            'unique_ips_limit_enabled': True,
            'unique_ips_limit_per_user': 10,
            'unique_ips_limit_time_window': 3600,
            # 'updating_name_disabled_for_users': 'some string',
            'usage_ping_enabled': False,
            # 'usage_ping_features_enabled': True,
            # 'user_deactivation_emails_enabled': False,
            'user_default_external': True,
            'user_default_internal_regex': 'some string',
            'user_oauth_applications': True,
            'user_show_add_ssh_key_message': False,
            # 'users_get_by_id_limit': 300,
            # 'users_get_by_id_limit_allowlist_raw': 'some string',
            'version_check_enabled': True,
            'web_ide_clientside_preview_enabled': True,
            # 'whats_new_variant': 0,
            # 'wiki_page_max_content_bytes': 52428800,
        })
        self.assertEqual(obj.errors(), [])
        if self.gitlab_version(lt='3'):
            return
        obj.process()
        self.assertTrue(cass.all_played)
